import re
from typing import List, Union, Callable

from ngram_graphs.TextGraph.IGraphTextGraph import IGraphTextGraph
from ngram_graphs.TextGraph.NetworkxTextGraph import NetworkxTextGraph
from ngram_graphs.utils import find_edge


class Generator:
    def __init__(self, n: int, kind: str = 'networkx'):
        self.n: int = n
        self.kind: str = kind
        self.__validate_generator()

    def __validate_generator(self):
        if self.n < 1:
            raise ValueError(f'{self.n} is not a valid value for n')
        if self.kind not in ['igraph', 'networkx']:
            raise ValueError(f'{self.kind} isn\'t a supported graph library. Please use "igraph" or "networkx"')

    def __get_chargrams(self, doc: str):
        return [doc[x:x + self.n:1] for x in range(len(doc) - self.n + 1)]

    def __get_regexngrams(self, doc: str, pattern: str):
        try:
            pattern = re.compile(pattern)
            doc = pattern.split(doc)
        except re.error as e:
            print(f"Error with regex '{pattern}'")
            raise re.error(e)
        grams = [doc[x:x + self.n:1] for x in range(len(doc) - self.n + 1)]
        return ['_'.join(g) for g in grams]

    def __get_wordngrams(self, doc: str, sep: str):
        if not sep:
            doc = doc.split()
        elif len(sep) == 1:
            doc = doc.split(sep)
        else:
            raise ValueError(f'{sep} is not a valid single char seperator.\n'
                             'If you meant to use a regex, set wordgram=False in the parameters')

        wordgrams = [doc[x:x + self.n:1] for x in range(len(doc) - self.n + 1)]
        return ['_'.join(wg) for wg in wordgrams]

    def __run_custom_function(self, doc: str, f: Callable[[str], List[str]]):
        return f(doc)

    def __generate_igraph(self, ngrams: List, weight: float):
        graph = IGraphTextGraph()
        for gram in ngrams:
            if gram not in [v["name"] for v in graph.vs]:
                graph.add_vertex(name=gram)

        for n1, n2 in zip(ngrams, ngrams[1:]):
            e = find_edge(graph, n1, n2)
            if e:
                e["weight"] += weight
            else:
                graph.add_edge(n1, n2, name=f'{n1} {n2}', weight=weight)
        return graph

    def __generate_networkx(self, ngrams: List, weight: float):
        graph = NetworkxTextGraph()
        for gram in ngrams:
            if gram not in graph.nodes:
                graph.add_node(gram)

        for n1, n2 in zip(ngrams, ngrams[1:]):
            e = find_edge(graph, n1, n2)
            if e:
                e["weight"] += weight
            else:
                graph.add_edge(n1, n2, name=f'{n1} {n2}', weight=weight)
        return graph

    def __generate_graph(self, ngrams: List, weight: float):
        if self.kind == 'igraph':
            return self.__generate_igraph(ngrams, weight)
        elif self.kind == 'networkx':
            return self.__generate_networkx(ngrams, weight)

    def __generate_text_graph(self, doc: str, weight: float, token: bool, sep: str):
        if token:
            graph = self.__generate_graph(self.__get_wordngrams(doc, sep), weight)
        elif isinstance(sep, str):
            graph = self.__generate_graph(self.__get_regexngrams(doc, sep), weight)
        elif isinstance(sep, Callable):
            graph = self.__generate_graph(self.__run_custom_function(doc, sep), weight)
        else:
            graph = self.__generate_graph(self.__get_chargrams(doc), weight)
        return graph

    def generate_text_graphs(self, documents: List[str], weight: float=1.0, wordgram: bool=False, sep: Union[str, Callable[[str], List[str]]]=None):
        graphs = [self.__generate_text_graph(doc, weight, wordgram, sep) for doc in documents]
        if len(graphs) and isinstance(graphs[0], IGraphTextGraph):
            graphs = [g for g in graphs if len(g.es)]
        elif len(graphs) and isinstance(graphs[0], NetworkxTextGraph):
            graphs = [g for g in graphs if len(g.edges)]
        return graphs

    def set_n(self, n):
        self.n = n
        self.__validate_generator()



from copy import copy
from typing import List, Union

from ngram_graphs.TextGraph.IGraphTextGraph import IGraphTextGraph, find_node_name
from ngram_graphs.TextGraph.NetworkxTextGraph import NetworkxTextGraph


def find_node(graph, node_idx):
    return graph.vs[node_idx]


def find_edge(graph, source, target):
    if isinstance(graph, IGraphTextGraph):
        for edge in graph.es:
            if (source, target) == (find_node_name(graph, edge.source), find_node_name(graph, edge.target)):
                return edge
        return None
    elif isinstance(graph, NetworkxTextGraph):
        if graph.has_edge(source, target):
            return graph.edges[source, target]
        return None


def generate_model_graph(graphs: List, lr: float = 0.5) -> Union[IGraphTextGraph, NetworkxTextGraph]:
    if not graphs:
        raise IndexError('Empty list of graph')
    model = copy(graphs[0])
    for graph in graphs[1:]:
        model.update(graph, lr)
    return model



from typing import Union

from ngram_graphs.TextGraph.IGraphTextGraph import IGraphTextGraph
from ngram_graphs.TextGraph.NetworkxTextGraph import NetworkxTextGraph


def size_similarity(g1: Union[IGraphTextGraph, NetworkxTextGraph], g2: Union[IGraphTextGraph, NetworkxTextGraph]):
    if isinstance(g1, IGraphTextGraph) and isinstance(g2, IGraphTextGraph):
        s1 = len(g1.es)
        s2 = len(g2.es)
    elif isinstance(g1, NetworkxTextGraph) and isinstance(g2, NetworkxTextGraph):
        s1 = len(g1.edges)
        s2 = len(g2.edges)
    else:
        raise TypeError(f'Invalid Graph types\ng1 is a {type(g1)} and g2 is a {type(g2)}')

    if not s1 or not s2:
        return None

    return min(s1, s2) / max(s1, s2)


def containment_similarity(g1: Union[IGraphTextGraph, NetworkxTextGraph], g2: Union[IGraphTextGraph, NetworkxTextGraph]):
    if isinstance(g1, IGraphTextGraph) and isinstance(g2, IGraphTextGraph):
        s1 = len(g1.es)
        s2 = len(g2.es)
        if not s1 or not s2:
            return None
        common_edges = [e for e in g1.es if e["name"] in g2.es["name"]]

        return len(common_edges) / max(s1, s2)
    elif isinstance(g1, NetworkxTextGraph) and isinstance(g2, NetworkxTextGraph):
        s1 = len(g1.edges)
        s2 = len(g2.edges)
        if not s1 or not s2:
            return None
        common_edges = [e for e in g1.edges if e in g2.edges]

        return len(common_edges) / max(s1, s2)
    else:
        raise TypeError(f'Invalid Graph types\ng1 is a {type(g1)} and g2 is a {type(g2)}')


def value_ratio(e1, e2):
    w1 = e1["weight"]
    w2 = e2["weight"]

    return min(w1, w2) / max(w1, w2)


def value_similarity(g1: Union[IGraphTextGraph, NetworkxTextGraph], g2: Union[IGraphTextGraph, NetworkxTextGraph]):
    if isinstance(g1, IGraphTextGraph) and isinstance(g2, IGraphTextGraph):
        s1 = len(g1.es)
        s2 = len(g2.es)
        common_edges = [(e1, e2) for e1 in g1.es for e2 in g2.es if e1["name"] == e2["name"]]

        return sum(value_ratio(e[0], e[1]) for e in common_edges) / max(s1, s2)

    elif isinstance(g1, NetworkxTextGraph) and isinstance(g2, NetworkxTextGraph):
        e1 = set(g1.edges)
        e2 = set(g2.edges)

        s1 = len(e1)
        s2 = len(e2)

        common_edges = e1 & e2
        common_edges = [(g1.get_edge_data(t[0], t[1]), g2.get_edge_data(t[0], t[1])) for t in common_edges]
        return sum(value_ratio(e[0], e[1]) for e in common_edges) / max(s1, s2)
    else:
        raise TypeError(f'Invalid Graph types\ng1 is a {type(g1)} and g2 is a {type(g2)}')


def normalized_value_similarity(g1: Union[IGraphTextGraph, NetworkxTextGraph], g2: Union[IGraphTextGraph, NetworkxTextGraph]):
    s = size_similarity(g1, g2)
    v = value_similarity(g1, g2)
    if not s:
        return None
    return v / s



from igraph import Graph


def find_node_name(graph, node_idx):
    return graph.vs[node_idx]["name"]


class IGraphTextGraph(Graph):
    def __init__(self):
        super().__init__(directed=True)

    def __copy__(self):
        g = IGraphTextGraph()
        for v in self.vs:
            g.add_vertex(v)
        for e in self.es:
            source = find_node_name(self, e.source)
            target = find_node_name(self, e.target)
            self.add_edge(source, target, weight=e["weight"], name=source + ' ' + target)
        return g

    @staticmethod
    def __calc_new_weight(s_edge, weight, learning_factor) -> float:
        return s_edge["weight"] + ((weight - s_edge["weight"]) * learning_factor)

    def __add_unknown_vertices(self, other):
        for vertex in other.vs:
            if vertex["name"] not in self.vs["name"]:
                self.add_vertex(name=vertex["name"])

    def __add_unknown_edge(self, other, o_edge, learning_factor):
        source = find_node_name(other, o_edge.source)
        target = find_node_name(other, o_edge.target)
        self.add_edge(source, target, weight=o_edge["weight"] * learning_factor, name=source + ' ' + target)

    def __update_edges(self, other: 'IGraphTextGraph', learning_factor: float):
        for o_edge in other.es:
            s_edge = next(iter([e for e in self.es if e["name"] == o_edge["name"]]), None)
            if s_edge is not None:
                s_edge["weight"] = self.__calc_new_weight(s_edge, o_edge["weight"], learning_factor)
            else:
                self.__add_unknown_edge(other, o_edge, learning_factor)
        for s_edge in self.es:
            if s_edge["name"] not in other.es["name"]:
                s_edge["weight"] = self.__calc_new_weight(s_edge, 0, learning_factor)

    def update(self, other: 'IGraphTextGraph', learning_factor):
        self.__add_unknown_vertices(other)
        self.__update_edges(other, learning_factor)



from networkx import DiGraph


def find_node_name(graph, name):
    try:
        return graph.nodes[name]
    except KeyError:
        return None


class NetworkxTextGraph(DiGraph):
    def __init__(self):
        super().__init__()

    def __copy__(self):
        g = NetworkxTextGraph()
        g.add_nodes_from(self.nodes(data=True))
        g.add_edges_from(self.edges(data=True))
        return g

    @staticmethod
    def __calc_new_weight(s_edge, weight: float, learning_factor: float) -> float:
        return s_edge["weight"] + ((weight - s_edge["weight"]) * learning_factor)

    def __add_unknown_vertices(self, other):
        for vertex in other.nodes:
            self.add_node(vertex)

    def __add_unknown_edge(self, o_edge, learning_factor):
        source = o_edge[0]
        target = o_edge[1]
        self.add_edge(source, target, weight=o_edge[2]["weight"] * learning_factor, name=source + ' ' + target)

    def __update_edges(self, other: 'NetworkxTextGraph', learning_factor: float):
        for o_edge in other.edges(data=True):
            edge_tuple = (o_edge[0], o_edge[1])
            if self.has_edge(o_edge[0], o_edge[1]):
                self.edges()[edge_tuple].update({"weight": self.__calc_new_weight(self.edges()[edge_tuple], o_edge[2]["weight"], learning_factor)})
            else:
                self.__add_unknown_edge(o_edge, learning_factor)
        for s_edge in self.edges(data=True):
            edge_tuple = (s_edge[0], s_edge[1])
            if edge_tuple not in other.edges:
                self.edges[edge_tuple].update({"weight": self.__calc_new_weight(s_edge[2], 0, learning_factor)})

    def update(self, other: 'NetworkxTextGraph', learning_factor):
        self.__add_unknown_vertices(other)
        self.__update_edges(other, learning_factor)



from unittest import TestCase

import ngram_graphs


class TestGraphs(TestCase):
    def setUp(self):
        n = 2
        self.data = ['head of marketing', 'marketing manager', 'marketing consultant', 'marketing']
        self.test = ['marketing manager', 'market analysis', 'construction worker']

        generator = ngram_graphs.Generator(n)

        self.data_graphs = generator.generate_text_graphs(self.data)
        self.test_graphs = generator.generate_text_graphs(self.test)

    def test_graph_sizes(self):
        self.assertEqual(len(self.data_graphs), 4)
        self.assertEqual(len(self.test_graphs), 3)

        self.assertEqual(len(self.data_graphs[0].vs), 16)
        self.assertEqual(len(self.data_graphs[0].es), 15)

    def test_comparisons(self):
        g1 = ngram_graphs.generate_model_graph(self.data_graphs)

        for idx, g2 in enumerate(self.test_graphs):
            print(self.test[idx])
            print('\tsize_similarity :\t\t\t\t', ngram_graphs.size_similarity(g1, g2))
            print('\tcontainment_similarity :\t\t', ngram_graphs.containment_similarity(g1, g2))
            print('\tvalue_similarity :\t\t\t\t', ngram_graphs.value_similarity(g1, g2))
            print('\tnormalized_value_similarity :\t', ngram_graphs.normalized_value_similarity(g1, g2))




from .Generator import Generator
from ngram_graphs.TextGraph.IGraphTextGraph import IGraphTextGraph
from ngram_graphs.TextGraph.NetworkxTextGraph import NetworkxTextGraph
from .utils import generate_model_graph
from .comparison import size_similarity, value_similarity, normalized_value_similarity, containment_similarity



from setuptools import setup, find_packages

setup(name='ngram_graphs',
      version='0.1.13',
      description='An igraph or networkx based ngram graph library for NLP',
      long_description="""Library for ngram graph comparison based on papers from George Giannakopoulos and Vangelis Karkaletsis. The library uses igraph to create directed and weighted graphs from documents and provides means to compare them using the distance metrics defined in 'Giannakopoulos, George & Karkaletsis, Vangelis. (2009). N-gram graphs: Representing documents and document sets in summary system evaluation. '""",
      url='https://github.com/loginn/ngrams_graphs',
      author='loginn',
      classifiers=[
        'Programming Language :: Python :: 3.6',
        'Topic :: Text Processing :: Linguistic',
      ],
      keywords='ngrams graph NLP',
      author_email='vincent.kurta2@gmail.com',
      license='MIT',
      install_requires=['python-igraph', 'networkx'],
      packages=find_packages(),
      zip_safe=False,
      test_suite='nose.collector',
      tests_require=['nose'],
      )
